@@ -1,11 +1,12 @@
 from crystally.core.vectors import *
 import copy
-import crystally.core.constants as constants
+from crystally.core.constants import *
 from typing import Union as Union
 import warnings
 import numpy as np
+from collections import Counter, namedtuple
 
-__all__ = ["Atom", "Lattice", "concat_lattices"]
+__all__ = ["Atom", "Lattice", "concat_lattices", "generate_rdf"]
 
 
 class Atom:
@@ -119,37 +120,36 @@ class Lattice:
         """
         return np.cross(self.vectors[0], self.vectors[1]).dot(self.vectors[2])
 
-    def sort(self, key=None, tolerance=None):
-        """ Sort the lattice with a given key.
+    def sort(self, key=None, *, tolerance=None, ignore_element=False):
+        """ Sort the lattice in-place with a given key.
 
         :param key: sorting argument which is used to sort the lattice. The following aruments can be used:
 
                     1. "element" - will sort according to the atoms element names
                     2. "position" - will sort according to the atoms absolute distance to the lattice origin
-                    3. "sublattice" - will sort according to the atoms sublattice name
-                    4. :class:`Lattice` - this will sort this lattice according to a different lattice. Equal atoms in
-                       this and the other lattice should then have the same position in the atoms lists. Element and
-                       Sublattice of the atoms are compared for equality. The position is compared with a tolerance in
-                       angstrom. This means that if the atoms are closer together than the tolerance, their position
-                       is treated as equal.
+                    3. :class:`Lattice` - this will sort this lattice according to a different lattice. Atoms are
+                        compared for their position and element. The latter can be deactivated.
 
         :param tolerance: tolerance parameter which will be used when sorting this lattice with another lattice. If
                           None is specified the default value of constants.ABS_VEC_COMP_TOL is used.
-        :return: nothing
+
+        :param ignore_element: ignores the element when the lattice is sorted with a different lattice
+
+        :return: None
         """
         if key is None:
-            self.atoms.sort(key=lambda atom: (atom.element, atom.sublattice, self.distance(atom, [0.0, 0.0, 0.0])))
+            self.atoms.sort(key=lambda atom: (atom.element, self.distance(atom, [0.0, 0.0, 0.0])))
         elif key == "element":
             self.atoms.sort(key=lambda atom: atom.element)
         elif key == "position":
             self.atoms.sort(key=lambda atom: self.distance(atom, [0.0, 0.0, 0.0]))
         elif isinstance(key, Lattice):
-            self._sort_with_other_lattice(key, tolerance)
+            self._sort_with_other_lattice(key, tolerance, ignore_element)
         else:
             self.atoms.sort(key=key)
         return None
 
-    def find(self, position=None, tolerance=None):
+    def find(self, position=None, *, tolerance=None):
         """ Search the atom list for an atom at the specified location
 
         :param position: position of the atoms to get within a tolerance (can be specified)
@@ -157,7 +157,7 @@ class Lattice:
         :return: :class:`Atoms <.Atom>` or `None` if no atoms satisfy the criteria.
         """
 
-        tolerance = constants.ABS_VEC_COMP_TOL if not tolerance else tolerance
+        tolerance = ABS_VEC_COMP_TOL if not tolerance else tolerance
 
         for atom in self.atoms:
             if self.distance(atom, position) < tolerance:
@@ -186,7 +186,7 @@ class Lattice:
         position = self._prepare_position(position, False)
         return self.vectors.dot(np.array(position))
 
-    def to_fractional(self, position, periodic=False):
+    def to_fractional(self, position, *, periodic=False):
         """ Translates cartesian coordinates to fractional coordinates.
 
         :param position: cartesian coordinates
@@ -201,7 +201,7 @@ class Lattice:
         else:
             return frac_position
 
-    def get_in_radius(self, center: Union[np.array, list, Atom], max_radius, min_radius=0.0):
+    def get_in_radius(self, center: Union[list, Atom], max_radius: float, min_radius: float = 0.0):
         """ Get all :class:`.Atom` that are in range to a provided position.
 
         :param center: Search center in fractional coordinates. This can either be a coordinates as a list
@@ -217,7 +217,7 @@ class Lattice:
         atom_list.sort(key=lambda atom: self.distance(atom, center))
         return atom_list
 
-    def index(self, position, tolerance=1e-3):
+    def index(self, position, *, tolerance=1e-3):
         """ Get the index within the atom list of the :class:`.Atom` at the provided position
 
         :param position: Position or Atom whose index is searched for
@@ -238,7 +238,7 @@ class Lattice:
         [element_names.add(x.element) for x in self.atoms]
         return list(element_names)
 
-    def distance(self, position1, position2, periodic=True) -> float:
+    def distance(self, position1, position2, *, periodic=True) -> float:
         """ Get the distance between two points
 
         :param position1: fractional coordinate as list or atom
@@ -259,7 +259,7 @@ class Lattice:
         diff_vector_cartesian = diff_vector.dot(self.vectors)
         return np.sqrt(diff_vector_cartesian.dot(diff_vector_cartesian))
 
-    def angle(self, position1, center, position2, periodic=True) -> float:
+    def angle(self, position1, center, position2, *, periodic=True) -> float:
         """ Get the angle between two positions and a center.
 
         :param position1: fractional coordinate as list or atom
@@ -287,7 +287,7 @@ class Lattice:
         magn2 = np.sqrt(vec_cart2.dot(vec_cart2))
         return np.arccos(vec_cart1.dot(vec_cart2)/(magn1*magn2))
 
-    def polyhedron_volume(self, vertices: list, periodic=True) -> float:
+    def polyhedron_volume(self, vertices, *, periodic=True) -> float:
         """ Lets you calculate the volume of a convex polyhedron (a body with no dents).
 
         :param vertices: List of fractional coordinates representing the vertices of the polyhedron.
@@ -315,7 +315,7 @@ class Lattice:
             warnings.warn("Points do not form a convex polyhedron", UserWarning)
         return convex_hull.volume
 
-    def polyhedron_area(self, vertices, periodic=True) -> float:
+    def polyhedron_area(self, vertices, *, periodic=True) -> float:
         """ Lets you calculate the area of a convex polyhedron (a body with no dents).
 
         :param vertices: List of fractional coordinates representing the vertices of the polyhedron.
@@ -358,7 +358,7 @@ class Lattice:
         position = self._prepare_position(position, True)
 
         # check if center and position are identical
-        if self.distance(center, position) < constants.ABS_VEC_COMP_TOL:
+        if self.distance(center, position) < ABS_VEC_COMP_TOL:
             raise ValueError("Center and Position are too close together")
 
         # calculate the distance in fractional coordinates from center to position
@@ -388,12 +388,13 @@ class Lattice:
         rel_distance = abs_increase/distance
         return self.increase_distance_rel(center, position, rel_distance)
 
-    def diff(self, other, tolerance=None):
-        """ Find the atomic differences bewteen this and another lattice. Atoms are compared according to element
-            sublattice and position. Mind that the fractional position of this and the other lattice are converted
+    def diff(self, other, *, tolerance=None, ignore_element=False):
+        """ Find the atomic differences between this and another lattice. Atoms are compared according to element
+            and position. Mind that the fractional position of this and the other lattice are converted
             to cartesian coordinates with the vectors of this lattice.
 
         :param other: other :class:`.Lattice`.
+        :param ignore_element: Ignore elements of lattices during comparison. Must be set to True or False.
         :param tolerance: tolerance in angstrom with which the atoms position are compared.
         :return: tuple with two lists, the first list consists of the atoms that were in this lattice, but were not
                  found in the other lattice. The second list contains the atoms that were in the other lattice, but not
@@ -401,14 +402,14 @@ class Lattice:
         """
 
         if not tolerance:
-            tolerance = constants.ABS_VEC_COMP_TOL
+            tolerance = ABS_VEC_COMP_TOL
 
         self_not_found = []
         other_not_found = list(other.atoms)
 
         for atom1 in self:
             for atom2_id, atom2 in enumerate(other_not_found):
-                if self._compare_atoms(atom1, atom2, tolerance):
+                if self._compare_atoms(atom1, atom2, tolerance, ignore_element):
                     del(other_not_found[atom2_id])
                     break
             else:
@@ -416,28 +417,29 @@ class Lattice:
 
         return self_not_found, other_not_found
 
-    def _sort_with_other_lattice(self, other, tolerance=None):
+    def _sort_with_other_lattice(self, other, tolerance=None, ignore_element=False):
         if tolerance is None:
-            tolerance = constants.ABS_VEC_COMP_TOL
+            tolerance = ABS_VEC_COMP_TOL
 
         old_order = list(self.atoms)
         new_order = []
         for atom1 in other:
             for atom2_id, atom2 in enumerate(old_order):
-                if self._compare_atoms(atom1, atom2, tolerance):
+                if self._compare_atoms(atom1, atom2, tolerance, ignore_element):
                     new_order.append(atom2)
                     del(old_order[atom2_id])
         new_order += old_order
         self.atoms = new_order
 
-    def _compare_atoms(self, atom1, atom2, dist_tolerance):
-        if atom1.element != atom2.element:
+    def _compare_atoms(self, atom1, atom2, dist_tolerance, ignore_element):
+        if atom1.element != atom2.element and ignore_element is False:
             return False
         if self.distance(atom1, atom2) > dist_tolerance:
             return False
         return True
 
-    def _prepare_position(self, position, periodic):
+    @staticmethod
+    def _prepare_position(position, periodic):
         """ Helper function to prepare positions, which were given as arguments.
 
         :param position: Either an Atom or a Vector (e.g. list / np.array / PeriodicVector)
@@ -482,7 +484,7 @@ def concat_lattices(lattice1: Lattice, lattice2: Lattice, direction: int):
     for i in range(lattice1.vectors.shape[0]):
         if i == direction:
             continue
-        if not np.allclose(lattice1.vectors[i], lattice2.vectors[i], atol=1e-10):
+        if not np.allclose(lattice1.vectors[i], lattice2.vectors[i], atol=FLOAT_ROUNDING):
             raise ValueError("shape of lattices does not match: "
                              "lattice1: {} lattice2: {}".format(str(lattice1.vectors[i]), lattice2.vectors[i]))
     new_lattice = Lattice(vectors=lattice1.vectors, atoms=[])
@@ -505,3 +507,64 @@ def concat_lattices(lattice1: Lattice, lattice2: Lattice, direction: int):
     return new_lattice
 
 
+def generate_rdf(lattice: Lattice, max_radius, *, min_radius=0.001, binning_fineness=1e-10):
+    """ Generate a rdf for a given :class:`.Lattice`. The neighbors are grouped by their element and distance.
+
+    :param lattice: the :class:`.Lattice` on which the rdf is performed
+    :param max_radius: maximum radius to which neighbors are searched for
+    :param min_radius: minimum radius from which neighbors are searched
+    :param binning_fineness: fineness of binning for the distance of neighbors
+    :return: list of namedtuples with the following fields: center, neighbor, distance, average_occurrence
+    """
+
+    def calculate_distances(center, neighbors):
+        diff_vectors = neighbors - center
+        diff_vectors = diff_vectors - (np.abs(diff_vectors) > 0.5) * np.sign(diff_vectors)
+        diff_vectors = np.dot(diff_vectors, lattice.vectors)
+        distances = np.linalg.norm(diff_vectors, axis=1)
+        return distances
+
+    def generate_position_array():
+        position_list = []
+        for atom in lattice:
+            position_list.append(atom.position.value)
+        positions = np.stack(position_list)
+        return positions
+
+    def count_elements():
+        element_count = Counter()
+        for atom in lattice:
+            element_count[atom.element] += 1
+        return element_count
+
+    def index_of_neighbors_in_range(distances):
+        return ((min_radius <= distances) & (distances <= max_radius)).nonzero()[0]
+
+    def bin_distance(dist):
+        dist = round(dist, FLOAT_PRECISION)
+        dist = dist // binning_fineness * binning_fineness
+        return round(dist, FLOAT_PRECISION)
+
+    positions = generate_position_array()
+    element_count = count_elements()
+    rdf_entry = namedtuple("rdf_entry", ["center", "neighbor", "distance", "average_occurrence"])
+
+    # generate rdf
+    rdf = Counter()
+    for atom in lattice:
+        distance_to_neighbors = calculate_distances(atom.position.value, positions)
+        for index in index_of_neighbors_in_range(distance_to_neighbors):
+            distance, neighbor = distance_to_neighbors[index], lattice[index]
+            distance = bin_distance(distance)
+            mean_addition = 1 / element_count[atom.element]
+            key = (atom.element, neighbor.element, distance)
+            rdf[key] += mean_addition
+
+    # sort rdf
+    sorted_rdf_keys = sorted(rdf.keys(), key=lambda x: (x[0], x[1], x[2]))
+    sorted_rdf = []
+    for key in sorted_rdf_keys:
+        entry = rdf_entry(center=key[0], neighbor=key[1], distance=key[2], average_occurrence=rdf[key])
+        sorted_rdf.append(entry)
+
+    return sorted_rdf
